/**CFile***********************************************************************

  FileName    [cuddAPI.c]

  PackageName [cudd]

  Synopsis    [Application interface functions.]

  Description [External procedures included in this module:
		<ul>
	        <li> Cudd_addNewVar()
		<li> Cudd_addNewVarAtLevel()
      		<li> Cudd_bddNewVar()
		<li> Cudd_bddNewVarAtLevel()
		<li> Cudd_addIthVar()
		<li> Cudd_bddIthVar()
		<li> Cudd_zddIthVar()
		<li> Cudd_zddVarsFromBddVars()
		<li> Cudd_addConst()
		<li> Cudd_AutodynEnable()
		<li> Cudd_AutodynDisable()
		<li> Cudd_ReorderingStatus()
		<li> Cudd_AutodynEnableZdd()
		<li> Cudd_AutodynDisableZdd()
		<li> Cudd_ReorderingStatusZdd()
		<li> Cudd_zddRealignmentEnabled()
		<li> Cudd_zddRealignEnable()
		<li> Cudd_zddRealignDisable()
		<li> Cudd_bddRealignmentEnabled()
		<li> Cudd_bddRealignEnable()
		<li> Cudd_bddRealignDisable()
		<li> Cudd_ReadOne()
		<li> Cudd_ReadZddOne()
		<li> Cudd_ReadZero()
		<li> Cudd_ReadLogicZero()
		<li> Cudd_ReadPlusInfinity()
		<li> Cudd_ReadMinusInfinity()
		<li> Cudd_ReadBackground()
		<li> Cudd_SetBackground()
		<li> Cudd_ReadCacheSlots()
		<li> Cudd_ReadCacheLookUps()
		<li> Cudd_ReadCacheHits()
		<li> Cudd_ReadMinHit()
		<li> Cudd_SetMinHit()
		<li> Cudd_ReadLooseUpTo()
		<li> Cudd_SetLooseUpTo()
		<li> Cudd_ReadMaxCache()
		<li> Cudd_SetMaxCache()
		<li> Cudd_ReadMaxCacheHard()
		<li> Cudd_SetMaxCacheHard()
		<li> Cudd_ReadSize()
		<li> Cudd_ReadSlots()
		<li> Cudd_ReadKeys()
		<li> Cudd_ReadDead()
		<li> Cudd_ReadMinDead()
		<li> Cudd_ReadReorderings()
		<li> Cudd_ReadSiftMaxVar()
		<li> Cudd_SetSiftMaxVar()
		<li> Cudd_ReadMaxGrowth()
		<li> Cudd_SetMaxGrowth()
		<li> Cudd_ReadTree()
		<li> Cudd_SetTree()
		<li> Cudd_FreeTree()
		<li> Cudd_ReadZddTree()
		<li> Cudd_SetZddTree()
		<li> Cudd_FreeZddTree()
                <li> Cudd_NodeReadIndex()
		<li> Cudd_ReadPerm()
		<li> Cudd_ReadInvPerm()
		<li> Cudd_ReadVars()
		<li> Cudd_ReadEpsilon()
		<li> Cudd_SetEpsilon()
		<li> Cudd_ReadGroupCheck()
		<li> Cudd_SetGroupcheck()
		<li> Cudd_GarbageCollectionEnabled()
		<li> Cudd_EnableGarbageCollection()
		<li> Cudd_DisableGarbageCollection()
		<li> Cudd_DeadAreCounted()
		<li> Cudd_TurnOnCountDead()
		<li> Cudd_TurnOffCountDead()
		<li> Cudd_ReadRecomb()
		<li> Cudd_SetRecomb()
		<li> Cudd_ReadSymmviolation()
		<li> Cudd_SetSymmviolation()
		<li> Cudd_ReadArcviolation()
		<li> Cudd_SetArcviolation()
		<li> Cudd_ReadPopulationSize()
		<li> Cudd_SetPopulationSize()
		<li> Cudd_ReadNumberXovers()
		<li> Cudd_SetNumberXovers()
		<li> Cudd_ReadMemoryInUse()
		<li> Cudd_PrintInfo()
		<li> Cudd_ReadPeakNodeCount()
		<li> Cudd_ReadNodeCount()
		<li> Cudd_zddReadNodeCount()
		<li> Cudd_AddHook()
		<li> Cudd_RemoveHook()
		</ul>]

  SeeAlso     []

  Author      [Fabio Somenzi]

  Copyright   [This file was created at the University of Colorado at
  Boulder.  The University of Colorado at Boulder makes no warranty
  about the suitability of this software for any purpose.  It is
  presented on an AS IS basis.]

******************************************************************************/

#include    "util.h"
#include    "cuddInt.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

#ifndef lint
static char rcsid[] DD_UNUSED = "$Id: cuddAPI.c,v 1.16 1997/04/13 10:15:58 fabio Exp $";
#endif

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/


/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/
#include    "message.h"
/**Function********************************************************************

  Synopsis    [Returns a new BDD variable.]

  Description [Creates a new BDD variable.  The new variable has an
  index equal to the largest previous index plus 1.  Returns a
  pointer to the new variable if successful; NULL otherwise.]

  SideEffects [None]

  SeeAlso     [Cudd_addNewVar Cudd_bddIthVar Cudd_bddNewVarAtLevel]

******************************************************************************/
DdNode *
Cudd_bddNewVar(dd)
DdManager *dd;
{
  DdNode* res;

  if ((unsigned int) dd->size >= CUDD_MAXINDEX - 1) return(NULL);
  if ( dd->size >= dd->maxSize)
    {
      ddResizeTablePar(dd,dd->size);
      ddResizeVars(dd);
    }

  /* all permutations are predefined to be identity */
  res=cuddUserNodeInter(dd,DdAssignNewVar(dd));
  if(res==NULL)
    return(NULL);
  res->ref=DD_MAXREF;
  dd->size++;
  return res;

} /* end of Cudd_bddNewVar */

/**Function********************************************************************

  Synopsis    [Returns the BDD variable with index i.]

  Description [Retrieves the BDD variable with index i if it already
  exists, or creates a new BDD variable.  Returns a pointer to the
  variable if successful; NULL otherwise.]

  SideEffects [None]

  SeeAlso     [Cudd_bddNewVar Cudd_addIthVar Cudd_bddNewVarAtLevel]

******************************************************************************/

DdNode *
Cudd_bddIthVar(dd,index)
DdManager *dd;
int index;
{
  int i;
    DdNode* res;

    if ( index >= CUDD_MAXINDEX - 1) return(NULL);
    if ( index >= dd->size)
      {
	if ( index >= dd->maxSize)
	  {
	    ddResizeTablePar(dd,index);
	    ddResizeVars(dd);
	  }
	for(i=dd->size;i<=index;i++)
	  dd->vars[i]=Cudd_bddNewVar(dd);
      }
    res = dd->vars[index];

    return(res);

} /* end of Cudd_bddIthVar */

/**Function********************************************************************

  Synopsis    [Enables automatic dynamic reordering of BDDs and ADDs.]

  Description [Enables automatic dynamic reordering of BDDs and
  ADDs. Parameter method is used to determine the method used for
  reordering. If CUDD_REORDER_SAME is passed, the method is
  unchanged.]

  SideEffects [None]

  SeeAlso     [Cudd_AutodynDisable Cudd_ReorderingStatus
  Cudd_AutodynEnableZdd]

******************************************************************************/
void
Cudd_AutodynEnable(unique,method)
DdManager *unique;
int method;
{
    unique->autoDyn = 1;
    if (method != CUDD_REORDER_SAME) {
	unique->autoMethod = method;
    }
    return;

} /* end of Cudd_AutodynEnable */


/**Function********************************************************************

  Synopsis    [Disables automatic dynamic reordering.]

  Description []

  SideEffects [None]

  SeeAlso     [Cudd_AutodynEnable Cudd_ReorderingStatus
  Cudd_AutodynDisableZdd]

******************************************************************************/
void
Cudd_AutodynDisable(unique)
DdManager *unique;
{
    unique->autoDyn = 0;
    return;

} /* end of Cudd_AutodynDisable */


/**Function********************************************************************

  Synopsis    [Reports the status of automatic dynamic reordering of BDDs
  and ADDs.]

  Description [Reports the status of automatic dynamic reordering of
  BDDs and ADDs. Parameter method is set to the reordering method
  currently selected. Returns 1 if automatic reordering is enabled; 0
  otherwise.]

  SideEffects [Parameter method is set to the reordering method currently
  selected.]

  SeeAlso     [Cudd_AutodynEnable Cudd_AutodynDisable
  Cudd_ReorderingStatusZdd]

******************************************************************************/
int
Cudd_ReorderingStatus(unique,method)
DdManager *unique;
int *method;
{
    *method = unique->autoMethod;
    return(unique->autoDyn);

} /* end of Cudd_ReorderingStatus */

/**Function********************************************************************

  Synopsis    [Returns the one constant of the manager.]

  Description [Returns the one constant of the manager. The one
  constant is common to ADDs and BDDs.]

  SideEffects [None]

  SeeAlso [Cudd_ReadZero Cudd_ReadLogicZero Cudd_ReadZddOne]

******************************************************************************/
DdNode*
Cudd_ReadOne(dd)
DdManager *dd;
{
    return(DD_ONE(dd));

} /* end of Cudd_ReadOne */

/**Function********************************************************************

  Synopsis    [Returns the logic zero constant of the manager.]

  Description [Returns the zero constant of the manager. The logic zero
  constant is the complement of the one constant, and is distinct from
  the arithmetic zero.]

  SideEffects [None]

  SeeAlso [Cudd_ReadOne Cudd_ReadZero]

******************************************************************************/
DdNode *
Cudd_ReadLogicZero(dd)
DdManager *dd;
{
    return(Cudd_Not(DD_ONE(dd)));

} /* end of Cudd_ReadLogicZero */

/**Function********************************************************************

  Synopsis    [Reads the number of slots in the cache.]

  Description []

  SideEffects [None]

******************************************************************************/
unsigned int
Cudd_ReadCacheSlots(dd)
DdManager *dd;
{
    return(dd->cacheSlots);

} /* end of Cudd_ReadCacheSlots */


/**Function********************************************************************

  Synopsis    [Returns the number of cache look-ups.]

  Description [Returns the number of cache look-ups. This number is
  approximate and is reinitialized by the package whenever the cache
  is resized.]

  SideEffects [None]

  SeeAlso     [Cudd_ReadCacheHits]

******************************************************************************/
unsigned int
Cudd_ReadCacheLookUps(dd)
DdManager *dd;
{
    return(dd->cacheLookUps);

} /* end of Cudd_ReadCacheLookUps */


/**Function********************************************************************

  Synopsis    [Returns the number of cache hits. This number is
  reinitialized by the package whenever the cache is resized.]

  Description []

  SideEffects [None]

  SeeAlso     [Cudd_ReadCacheLookUps]

******************************************************************************/
unsigned int
Cudd_ReadCacheHits(dd)
DdManager *dd;
{
    return(dd->cacheHits);

} /* end of Cudd_ReadCacheHits */


/**Function********************************************************************

  Synopsis    [Reads the minHit field of the manager.]

  Description []

  SideEffects [None]

  SeeAlso     [Cudd_SetMinHit]

******************************************************************************/
unsigned int
Cudd_ReadMinHit(dd)
DdManager *dd;
{
    return(dd->minHit);

} /* end of Cudd_ReadMinHit */


/**Function********************************************************************

  Synopsis    [Sets the minHit parameter of the manager.]

  Description [Sets the minHit parameter of the manager. This
  parameter controls the resizing of the cache. If the hit ratio is
  larger than minHit/100, and the cache is not already too large, then
  its size is doubled.]

  SideEffects [None]

  SeeAlso     [Cudd_ReadMinHit]

******************************************************************************/
void
Cudd_SetMinHit(dd,hr)
DdManager *dd;
unsigned int hr;
{
    dd->minHit = hr;

} /* end of Cudd_SetMinHit */


/**Function********************************************************************

  Synopsis    [Reads the looseUpTo parameter of the manager.]

  Description []

  SideEffects [None]

  SeeAlso     [Cudd_SetLooseUpTo Cudd_ReadMinHit Cudd_ReadMinDead]

******************************************************************************/
unsigned int
Cudd_ReadLooseUpTo(dd)
DdManager *dd;
{
    return(dd->looseUpTo);

} /* end of Cudd_ReadLooseUpTo */


/**Function********************************************************************

  Synopsis    [Sets the looseUpTo parameter of the manager.]

  Description [Sets the looseUpTo parameter of the manager. This
  parameter of the manager controls the threshold beyond which no fast
  growth of the unique table is allowed. The threshold is given as a
  number of slots.]

  SideEffects [None]

  SeeAlso     [Cudd_ReadLooseUpTo Cudd_SetMinHit]

******************************************************************************/
void
Cudd_SetLooseUpTo(dd,lut)
DdManager *dd;
unsigned int lut;
{
    dd->looseUpTo = lut;

} /* end of Cudd_SetLooseUpTo */


/**Function********************************************************************

  Synopsis    [Reads the maxCache parameter of the manager.]

  Description []

  SideEffects [None]

  SeeAlso     [Cudd_SetMaxCache Cudd_ReadMaxCache]

******************************************************************************/
unsigned int
Cudd_ReadMaxCache(dd)
DdManager *dd;
{
    return(dd->maxCache);

} /* end of Cudd_ReadMaxCache */


/**Function********************************************************************

  Synopsis    [Sets the maxCache parameter of the manager.]

  Description [Sets the maxCache parameter of the manager. This parameter
  is normally automatically determined by the package based on the total
  number of nodes and the parameter maxCacheHard. Applications should
  normally change maxCacheHard rather than maxCache if they want to
  explicitly control the trade-off of memory versus speed.]

  SideEffects [None]

  SeeAlso     [Cudd_ReadMaxCache Cudd_SetMaxCacheHard]

******************************************************************************/
void
Cudd_SetMaxCache(dd,mc)
DdManager *dd;
unsigned int mc;
{
    dd->maxCache = mc;

} /* end of Cudd_SetMaxCache */


/**Function********************************************************************

  Synopsis    [Reads the maxCacheHard parameter of the manager.]

  Description []

  SideEffects [None]

  SeeAlso     [Cudd_SetMaxCacheHard Cudd_ReadMaxCache]

******************************************************************************/
unsigned int
Cudd_ReadMaxCacheHard(dd)
DdManager *dd;
{
    return(dd->maxCacheHard);

} /* end of Cudd_ReadMaxCache */


/**Function********************************************************************

  Synopsis    [Sets the maxCacheHard parameter of the manager.]

  Description [Sets the maxCacheHard parameter of the manager. The
  cache cannot grow larger than maxCacheHard entries. This parameter
  allows an application to control the trade-off of memory versus
  speed. If the value passed to this function is 0, the function
  determines a suitable maximum cache size based on the available memory.]

  SideEffects [None]

  SeeAlso     [Cudd_ReadMaxCacheHard Cudd_SetMaxCache]

******************************************************************************/
void
Cudd_SetMaxCacheHard(dd,mc)
DdManager *dd;
unsigned int mc;
{
    if (mc == 0) {
      assert(0);
      /*
	long datalimit = getSoftDataLimit();
	mc = datalimit / sizeof(DdCache) / 4;
	*/
    }
    dd->maxCacheHard = mc;

} /* end of Cudd_SetMaxCacheHard */


/**Function********************************************************************

  Synopsis    [Returns the number of BDD variables in existance.]

  Description []

  SideEffects [None]

  SeeAlso     [Cudd_ReadZddSize]

******************************************************************************/
int
Cudd_ReadSize(dd)
DdManager *dd;
{
    return(dd->size);

} /* end of Cudd_ReadSize */

/**Function********************************************************************

  Synopsis    [Returns the total number of slots of the unique table.]

  Description [Returns the total number of slots of the unique table.
  This number ismainly for diagnostic purposes.]

  SideEffects [None]

******************************************************************************/
unsigned int
Cudd_ReadSlots(dd)
DdManager *dd;
{
    return(dd->slots);

} /* end of Cudd_ReadSlots */


/**Function********************************************************************

  Synopsis    [Returns the number of nodes in the unique table.]

  Description [Returns the total number of nodes currently in the unique
  table, including the dead nodes.]

  SideEffects [None]

  SeeAlso     [Cudd_ReadDead]

******************************************************************************/
unsigned int
Cudd_ReadKeys(dd)
DdManager *dd;
{
    return(dd->keys);

} /* end of Cudd_ReadKeys */


/**Function********************************************************************

  Synopsis    [Returns the number of dead nodes in the unique table.]

  Description []

  SideEffects [None]

  SeeAlso     [Cudd_ReadKeys]

******************************************************************************/
unsigned int
Cudd_ReadDead(dd)
DdManager *dd;
{
    return(dd->dead);

} /* end of Cudd_ReadDead */


/**Function********************************************************************

  Synopsis    [Returns the number of times reordering has occurred.]

  Description [Returns the number of times reordering has occurred in the
  manager. The number includes both the calls to Cudd_ReduceHeap from
  the application program and those automatically performed by the
  package. However, calls that do not even initiate reordering are not
  counted. A call may not initiate reordering if there are fewer than
  minsize live nodes in the manager, or if CUDD_REORDER_NONE is specified
  as reordering method. The calls to Cudd_ShuffleHeap are not counted.]

  SideEffects [None]

  SeeAlso [Cudd_ReduceHeap]

******************************************************************************/
int
Cudd_ReadReorderings(dd)
DdManager *dd;
{
    return(dd->reorderings);

} /* end of Cudd_ReadReorderings */


/**Function********************************************************************

  Synopsis    [Reads the siftMaxVar parameter of the manager.]

  Description [Reads the siftMaxVar parameter of the manager. This
  parameter gives the maximum number of variables that will be sifted
  for each invocation of sifting.]

  SideEffects [None]

  SeeAlso     [Cudd_ReadSiftMaxSwap Cudd_SetSiftMaxVar]

******************************************************************************/
int
Cudd_ReadSiftMaxVar(dd)
DdManager *dd;
{
    return(dd->siftMaxVar);

} /* end of Cudd_ReadSiftMaxVar */


/**Function********************************************************************

  Synopsis    [Sets the siftMaxVar parameter of the manager.]

  Description [Sets the siftMaxVar parameter of the manager. This
  parameter gives the maximum number of variables that will be sifted
  for each invocation of sifting.]

  SideEffects [None]

  SeeAlso     [Cudd_SetSiftMaxSwap Cudd_ReadSiftMaxVar]

******************************************************************************/
void
Cudd_SetSiftMaxVar(dd,smv)
DdManager *dd;
int smv;
{
    dd->siftMaxVar = smv;

} /* end of Cudd_SetSiftMaxVar */


/**Function********************************************************************

  Synopsis    [Reads the siftMaxSwap parameter of the manager.]

  Description [Reads the siftMaxSwap parameter of the manager. This
  parameter gives the maximum number of swaps that will be attempted
  for each invocation of sifting. The real number of swaps may exceed
  the set limit because the package will always complete the sifting
  of the variable that causes the limit to be reached.]

  SideEffects [None]

  SeeAlso     [Cudd_ReadSiftMaxVar Cudd_SetSiftMaxSwap]

******************************************************************************/
int
Cudd_ReadSiftMaxSwap(dd)
DdManager *dd;
{
    return(dd->siftMaxSwap);

} /* end of Cudd_ReadSiftMaxSwap */


/**Function********************************************************************

  Synopsis    [Sets the siftMaxSwap parameter of the manager.]

  Description [Sets the siftMaxSwap parameter of the manager. This
  parameter gives the maximum number of swaps that will be attempted
  for each invocation of sifting. The real number of swaps may exceed
  the set limit because the package will always complete the sifting
  of the variable that causes the limit to be reached.]

  SideEffects [None]

  SeeAlso     [Cudd_SetSiftMaxVar Cudd_ReadSiftMaxSwap]

******************************************************************************/
void
Cudd_SetSiftMaxSwap(dd,sms)
DdManager *dd;
int sms;
{
    dd->siftMaxSwap = sms;

} /* end of Cudd_SetSiftMaxSwap */


/**Function********************************************************************

  Synopsis    [Reads the maxGrowth parameter of the manager.]

  Description [Reads the maxGrowth parameter of the manager. This parameter
  determines how much the number of nodes can grow during sifting of a
  variable. Overall, sifting never increases the size of the decision
  diagrams. This parameter only refers to intermediate results. A lower
  value will speed up sifting, possibly at the expense of quality.]

  SideEffects [None]

  SeeAlso     [Cudd_SetMaxGrowth]

******************************************************************************/
double
Cudd_ReadMaxGrowth(dd)
DdManager *dd;
{
    return(dd->maxGrowth);

} /* end of Cudd_ReadMaxGrowth */


/**Function********************************************************************

  Synopsis    [Sets the maxGrowth parameter of the manager.]

  Description [Sets the maxGrowth parameter of the manager. This parameter
  determines how much the number of nodes can grow during sifting of a
  variable. Overall, sifting never increases the size of the decision
  diagrams. This parameter only refers to intermediate results. A lower
  value will speed up sifting, possibly at the expense of quality.]

  SideEffects [None]

  SeeAlso     [Cudd_ReadMaxGrowth]

******************************************************************************/
void
Cudd_SetMaxGrowth(dd,mg)
DdManager *dd;
double mg;
{
    dd->maxGrowth = mg;

} /* end of Cudd_SetMaxGrowth */

/**Function********************************************************************

  Synopsis    [Returns the index of the node.]

  Description [Returns the index of the node. The node pointer can be
  either regular or complemented.]

  SideEffects [None]

  SeeAlso [Cudd_ReadIndex]

******************************************************************************/
unsigned int
Cudd_NodeReadIndex(node)
DdNode    *node;
{
    return((unsigned int) Cudd_Regular(node)->index);

} /* end of Cudd_NodeReadIndex */

/**Function********************************************************************

  Synopsis    [Returns the current position of the i-th variable in the
  order.]

  Description [Returns the current position of the i-th variable in the
  order. If the index is CUDD_MAXINDEX, returns CUDD_MAXINDEX; otherwise,
  if the index is out of bounds returns -1.]

  SideEffects [None]

  SeeAlso     [Cudd_ReadInvPerm Cudd_ReadPermZdd]

  Updated

******************************************************************************/
int
Cudd_ReadPerm(dd,i)
DdManager *dd;
int i;
{
    if (i == CUDD_MAXINDEX) return(CUDD_MAXINDEX);
    if (i < 0 || i >= dd->size) return(-1);
    return(dd->OrigPerm[i]);

} /* end of Cudd_ReadPerm */

/**Function********************************************************************

  Synopsis    [Returns the index of the variable currently in the i-th
  position of the order.]

  Description [Returns the index of the variable currently in the i-th
  position of the order. If the index is CUDD_MAXINDEX, returns
  CUDD_MAXINDEX; otherwise, if the index is out of bounds returns -1.]

  SideEffects [None]

  SeeAlso     [Cudd_ReadPerm Cudd_ReadInvPermZdd]

******************************************************************************/
int
Cudd_ReadInvPerm(dd,i)
DdManager *dd;
int i;
{
    if (i == CUDD_MAXINDEX) return(CUDD_MAXINDEX);
    if (i < 0 || i >= dd->size) return(-1);
    return(dd->InvOrigPerm[i]);

} /* end of Cudd_ReadInvPerm */

/**Function********************************************************************

  Synopsis    [Returns the i-th element of the vars array.]

  Description [Returns the i-th element of the vars array if it falls
  within the array bounds; NULL otherwise. If i is the index of an
  existing variable, this function produces the same result as
  Cudd_bddIthVar. However, if the i-th var does not exist yet,
  Cudd_bddIthVar will create it, whereas Cudd_ReadVars will not.]

  SideEffects [None]

  SeeAlso     [Cudd_bddIthVar]

******************************************************************************/
DdNode *
Cudd_ReadVars(dd,i)
DdManager *dd;
int i;
{
    if (i < 0 || i > dd->size) return(NULL);
    return(dd->vars[i]);

} /* end of Cudd_ReadVars */

/**Function********************************************************************

  Synopsis    [Reads the groupcheck parameter of the manager.]

  Description [Reads the groupcheck parameter of the manager. The
  groupcheck parameter determines the aggregation criterion in group
  sifting.]

  SideEffects [None]

  SeeAlso     [Cudd_SetGroupcheck]

******************************************************************************/
int
Cudd_ReadGroupcheck(dd)
DdManager *dd;
{
    return(dd->groupcheck);

} /* end of Cudd_ReadGroupCheck */


/**Function********************************************************************

  Synopsis    [Sets the parameter groupcheck of the manager to gc.]

  Description [Sets the parameter groupcheck of the manager to gc. The
  groupcheck parameter determines the aggregation criterion in group
  sifting.]

  SideEffects [None]

  SeeAlso     [Cudd_ReadGroupCheck]

******************************************************************************/
void
Cudd_SetGroupcheck(dd,gc)
DdManager *dd;
int gc;
{
    dd->groupcheck = gc;

} /* end of Cudd_SetGroupcheck */

/**Function********************************************************************

  Synopsis    [Tells whether garbage collection is enabled.]

  Description [Returns 1 if garbage collection is enabled; 0 otherwise.]

  SideEffects [None]

  SeeAlso     [Cudd_EnableGarbageCollection Cudd_DisableGarbageCollection]

******************************************************************************/
int
Cudd_GarbageCollectionEnabled(dd)
DdManager *dd;
{
    return(dd->gcEnabled);

} /* end of Cudd_GarbageCollectionEnabled */


/**Function********************************************************************

  Synopsis    [Enables garbage collection.]

  Description [Enables garbage collection. Garbage collection is
  initially enabled. Therefore it is necessary to call this function
  only if garbage collection has been explicitly disabled.]

  SideEffects [None]

  SeeAlso     [Cudd_DisableGarbageCollection Cudd_GarbageCollectionEnabled]

******************************************************************************/
void
Cudd_EnableGarbageCollection(dd)
DdManager *dd;
{
    dd->gcEnabled = 1;

} /* end of Cudd_EnableGarbageCollection */


/**Function********************************************************************

  Synopsis    [Disables garbage collection.]

  Description [Disables garbage collection. Garbage collection is
  initially enabled. This function may be called to disable it.
  However, garbage collection will still occur when a new node must be
  created and no memory is left, or when garbage collection is required
  for correctness. (E.g., before reordering.)]

  SideEffects [None]

  SeeAlso     [Cudd_EnableGarbageCollection Cudd_GarbageCollectionEnabled]

******************************************************************************/
void
Cudd_DisableGarbageCollection(dd)
DdManager *dd;
{
    dd->gcEnabled = 0;

} /* end of Cudd_DisableGarbageCollection */

/**Function********************************************************************

  Synopsis    [Returns the current value of the recombination parameter used
  in group sifting.]

  Description [Returns the current value of the recombination
  parameter used in group sifting. A larger (positive) value makes the
  aggregation of variables due to the second difference criterion more
  likely. A smaller (negative) value makes aggregation less likely.]

  SideEffects [None]

  SeeAlso     [Cudd_SetRecomb]

******************************************************************************/
int
Cudd_ReadRecomb(dd)
DdManager *dd;
{
    return(dd->recomb);

} /* end of Cudd_ReadRecomb */


/**Function********************************************************************

  Synopsis    [Sets the value of the recombination parameter used in group
  sifting.]

  Description [Sets the value of the recombination parameter used in
  group sifting. A larger (positive) value makes the aggregation of
  variables due to the second difference criterion more likely. A
  smaller (negative) value makes aggregation less likely. The default
  value is 0.]

  SideEffects [Changes the manager.]

  SeeAlso     [Cudd_ReadRecomb]

******************************************************************************/
void
Cudd_SetRecomb(dd,recomb)
DdManager *dd;
int recomb;
{
    dd->recomb = recomb;

} /* end of Cudd_SetRecomb */


/**Function********************************************************************

  Synopsis    [Returns the current value of the symmviolation parameter used
  in group sifting.]

  Description [Returns the current value of the symmviolation
  parameter. This parameter is used in group sifting to decide how
  many violations to the symmetry conditions <code>f10 = f01</code> or
  <code>f11 = f00</code> are tolerable when checking for aggregation
  due to extended symmetry. The value should be between 0 and 100. A
  small value causes fewer variables to be aggregated. The default
  value is 0.]

  SideEffects [None]

  SeeAlso     [Cudd_SetSymmviolation]

******************************************************************************/
int
Cudd_ReadSymmviolation(dd)
DdManager *dd;
{
    return(dd->symmviolation);

} /* end of Cudd_ReadSymmviolation */


/**Function********************************************************************

  Synopsis    [Sets the value of the symmviolation parameter used
  in group sifting.]

  Description [Sets the value of the symmviolation
  parameter. This parameter is used in group sifting to decide how
  many violations to the symmetry conditions <code>f10 = f01</code> or
  <code>f11 = f00</code> are tolerable when checking for aggregation
  due to extended symmetry. The value should be between 0 and 100. A
  small value causes fewer variables to be aggregated. The default
  value is 0.]

  SideEffects [Changes the manager.]

  SeeAlso     [Cudd_ReadSymmviolation]

******************************************************************************/
void
Cudd_SetSymmviolation(dd,symmviolation)
DdManager *dd;
int symmviolation;
{
    dd->symmviolation = symmviolation;

} /* end of Cudd_SetSymmviolation */


/**Function********************************************************************

  Synopsis    [Returns the current value of the arcviolation parameter used
  in group sifting.]

  Description [Returns the current value of the arcviolation
  parameter. This parameter is used in group sifting to decide how
  many arcs into <code>y</code> not coming from <code>x</code> are
  tolerable when checking for aggregation due to extended
  symmetry. The value should be between 0 and 100. A small value
  causes fewer variables to be aggregated. The default value is 0.]

  SideEffects [None]

  SeeAlso     [Cudd_SetArcviolation]

******************************************************************************/
int
Cudd_ReadArcviolation(dd)
DdManager *dd;
{
    return(dd->arcviolation);

} /* end of Cudd_ReadArcviolation */


/**Function********************************************************************

  Synopsis    [Sets the value of the arcviolation parameter used
  in group sifting.]

  Description [Sets the value of the arcviolation
  parameter. This parameter is used in group sifting to decide how
  many arcs into <code>y</code> not coming from <code>x</code> are
  tolerable when checking for aggregation due to extended
  symmetry. The value should be between 0 and 100. A small value
  causes fewer variables to be aggregated. The default value is 0.]

  SideEffects [None]

  SeeAlso     [Cudd_ReadArcviolation]

******************************************************************************/
void
Cudd_SetArcviolation(dd,arcviolation)
DdManager *dd;
int arcviolation;
{
    dd->arcviolation = arcviolation;

} /* end of Cudd_SetArcviolation */

/**Function********************************************************************

  Synopsis    [Reads the current size of the population used by the
  genetic algorithm for reordering.]

  Description [Reads the current size of the population used by the
  genetic algorithm for variable reordering. A larger population size will
  cause the genetic algorithm to take more time, but will generally
  produce better results. The default value is 0, in which case the
  package uses three times the number of variables as population size,
  with a maximum of 120.]

  SideEffects [None]

  SeeAlso     [Cudd_SetPopulationSize]

******************************************************************************/
int
Cudd_ReadPopulationSize(dd)
DdManager *dd;
{
    return(dd->populationSize);

} /* end of Cudd_ReadPopulationSize */


/**Function********************************************************************

  Synopsis    [Sets the size of the population used by the
  genetic algorithm for reordering.]

  Description [Sets the size of the population used by the
  genetic algorithm for variable reordering. A larger population size will
  cause the genetic algorithm to take more time, but will generally
  produce better results. The default value is 0, in which case the
  package uses three times the number of variables as population size,
  with a maximum of 120.]

  SideEffects [Changes the manager.]

  SeeAlso     [Cudd_ReadPopulationSize]

******************************************************************************/
void
Cudd_SetPopulationSize(dd,populationSize)
DdManager *dd;
int populationSize;
{
    dd->populationSize = populationSize;

} /* end of Cudd_SetPopulationSize */


/**Function********************************************************************

  Synopsis    [Reads the current number of crossovers used by the
  genetic algorithm for reordering.]

  Description [Reads the current number of crossovers used by the
  genetic algorithm for variable reordering. A larger number of crossovers will
  cause the genetic algorithm to take more time, but will generally
  produce better results. The default value is 0, in which case the
  package uses three times the number of variables as number of crossovers,
  with a maximum of 60.]

  SideEffects [None]

  SeeAlso     [Cudd_SetNumberXovers]

******************************************************************************/
int
Cudd_ReadNumberXovers(dd)
DdManager *dd;
{
    return(dd->numberXovers);

} /* end of Cudd_ReadNumberXovers */


/**Function********************************************************************

  Synopsis    [Sets the number of crossovers used by the
  genetic algorithm for reordering.]

  Description [Sets the number of crossovers used by the genetic
  algorithm for variable reordering. A larger number of crossovers
  will cause the genetic algorithm to take more time, but will
  generally produce better results. The default value is 0, in which
  case the package uses three times the number of variables as number
  of crossovers, with a maximum of 60.]

  SideEffects [None]

  SeeAlso     [Cudd_ReadNumberXovers]

******************************************************************************/
void
Cudd_SetNumberXovers(dd,numberXovers)
DdManager *dd;
int numberXovers;
{
    dd->numberXovers = numberXovers;

} /* end of Cudd_SetNumberXovers */

/**Function********************************************************************

  Synopsis    [Returns the memory in use by the manager measured in bytes.]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/
long
Cudd_ReadMemoryInUse(dd)
DdManager *dd;
{
    return(dd->memused);

} /* end of Cudd_ReadMemoryInUse */

/**Function********************************************************************

  Synopsis    [Prints out statistics and settings for a CUDD manager.]

  Description [Prints out statistics and settings for a CUDD manager.
  Returns 1 if successful; 0 otherwise.]

  SideEffects [None]

  SeeAlso     []

******************************************************************************/
int
Cudd_PrintInfo(dd,fp)
DdManager *dd;
FILE *fp;
{
    int retval;
    int autoMethod;

    /* Modifiable parameters. */
    retval = fprintf(fp,"**** CUDD modifiable parameters ****\n");
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Hard limit for cache size: %u\n",
		     Cudd_ReadMaxCacheHard(dd));
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Cache hit threshold for resizing: %u%%\n",
		     Cudd_ReadMinHit(dd));
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Garbage collection enabled: %s\n",
		     Cudd_GarbageCollectionEnabled(dd) ? "yes" : "no");
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Limit for fast unique table growth: %u\n",
		     Cudd_ReadLooseUpTo(dd));
    if (retval == EOF) return(0);
    retval = fprintf(fp,
		     "Maximum number of variables sifted per reordering: %d\n",
		     Cudd_ReadSiftMaxVar(dd));
    if (retval == EOF) return(0);
    retval = fprintf(fp,
		     "Maximum number of variable swaps per reordering: %d\n",
		     Cudd_ReadSiftMaxSwap(dd));
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Maximum growth while sifting a variable: %g\n",
		     Cudd_ReadMaxGrowth(dd));
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Dynamic reordering of BDDs enabled: %s\n",
		     Cudd_ReorderingStatus(dd,&autoMethod) ? "yes" : "no");
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Default BDD reordering method: %d\n", autoMethod);
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Group checking criterion: %d\n",
		     Cudd_ReadGroupcheck(dd));
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Recombination threshold: %d\n", Cudd_ReadRecomb(dd));
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Symmetry violation threshold: %d\n",
		     Cudd_ReadSymmviolation(dd));
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Arc violation threshold: %d\n",
		     Cudd_ReadArcviolation(dd));
    if (retval == EOF) return(0);
    retval = fprintf(fp,"GA population size: %d\n",
		     Cudd_ReadPopulationSize(dd));
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Number of crossovers for GA: %d\n",
		     Cudd_ReadNumberXovers(dd));
    if (retval == EOF) return(0);

    /* Non-modifiable parameters. */
    retval = fprintf(fp,"**** CUDD non-modifiable parameters ****\n");
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Memory in use: %ld\n", Cudd_ReadMemoryInUse(dd));
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Number of BDD variables: %d\n", dd->size);
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Number of cache entries: %u\n", dd->cacheSlots);
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Number of cache look-ups: %u\n", dd->cacheLookUps);
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Number of cache hits: %u\n", dd->totCachehits);
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Number of cache collisions: %u\n",
		     dd->cachecollisions);
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Number of cache insertions: %u\n", dd->cacheinserts);
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Number of cache deletions: %u\n", dd->cachedeletions);
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Soft limit for cache size: %u\n", dd->maxCache);
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Number of buckets in unique table: %u\n", dd->slots);
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Number of BDD and ADD nodes: %u\n", dd->keys);
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Number of dead BDD and ADD nodes: %u\n", dd->dead);
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Total number of nodes allocated: %d\n",
		     dd->allocated);
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Garbage collections so far: %d\n",
		     dd->garbageCollections);
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Reorderings so far: %d\n", dd->reorderings);
    if (retval == EOF) return(0);
    retval = fprintf(fp,"Next reordering threshold: %u\n", dd->nextDyn);
    if (retval == EOF) return(0);

    return(1);

} /* end of Cudd_PrintInfo */


