/**CFile***********************************************************************

  FileName    [cuddBddAbs.c]

  PackageName [cudd]

  Synopsis    [Quantification functions for BDDs.]

  Description [External procedures included in this module:
		<ul>
		<li> Cudd_bddExistAbstract()
		<li> Cudd_bddUnivAbstract()
		<li> Cudd_bddBooleanDiff()
		</ul>
	Internal procedures included in this module:
		<ul>
		<li> cuddBddExistAbstractRecur()
		<li> cuddBddXorExistAbstractRecur()
		<li> cuddBddBooleanDiffRecur()
		</ul>
	Static procedures included in this module:
		<ul>
		<li> bddCheckPositiveCube()
		</ul>
		]

  Author      [Fabio Somenzi]

  Copyright   [This file was created at the University of Colorado at
  Boulder.  The University of Colorado at Boulder makes no warranty
  about the suitability of this software for any purpose.  It is
  presented on an AS IS basis.]

******************************************************************************/

#include "util.h"
#include "cuddInt.h"
#include "message.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

#ifndef lint
static char rcsid[] DD_UNUSED = "$Id: cuddBddAbs.c,v 1.11 1997/04/13 17:29:14 fabio Exp $";
#endif

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static DdNodePar cuddBddExistsSimple(DdManager *manager,DdNodePar f, DdNodePar *cube, DdCacheHandle *handle);
static int bddCheckPositiveCube ARGS((DdManager *manager, DdNodePar cube));

/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Performs the recursive steps of Cudd_bddExistAbstract.]

  Description [Performs the recursive steps of Cudd_bddExistAbstract.
  Returns the BDD obtained by abstracting the variables
  of cube from f if successful; NULL otherwise. It is also used by
  Cudd_bddUnivAbstract.]

  SideEffects [None]

  SeeAlso     [Cudd_bddExistAbstract Cudd_bddUnivAbstract]

******************************************************************************/

void cuddBddExistAbstractRecurPre(manager,f, cube, from, caller)
DdManager *manager;
DdNodePar  f;
DdNodePar  cube;
int from;
DdCallTreeNode caller;
{
  int op=CUDD_BDD_EXISTS;
  DdCacheHandle handle;
  DdCallTreeNode Node;

  assert(from!=manager->Mypos); /* should no be call internally */

  /* Abstract a variables that does not appear in f. */
  if(!cuddLocal(manager,f.index))
    { /* remove all the local variables from cube and pass on the work */
      while(cube.index!=CUDD_MAXINDEX && cuddLocal(manager,cube.index))
	cube=Cudd_T_Par(manager,cube);
    }
    else 
      {
	while(f.index>cube.index)
	  cube=Cudd_T_Par(manager,cube);
      }

  if(cuddEqualPar(cube,DD_ONE_PAR(manager)))
    /* f does not depend on any of the variables */
    DdMessageReply(manager,from,f,caller);	  
  else {
    /* check cache */
    handle = cuddCacheCheck2Par(manager, op, f, cube);
    if(cuddCacheStatusPar(manager,handle)==CUDD_CACHE_HIT)
      DdMessageReply(manager,from,cuddCacheResult(manager,handle),caller);    
    else 
      {
	Node=DdCallTreeCreateNode_Propagate(manager,caller,from,handle);
	if(cuddCacheStatusPar(manager,handle)==CUDD_CACHE_IN_PROGRESS)
	  {
	    cuddCacheWait(manager,handle,Node);
	    /* cannot just use caller as it may be on another machine */
	  }
	else if(cuddCacheStatusPar(manager,handle)==CUDD_CACHE_MISS)
	  {
	    DdWorkCreate2(manager,op,f,cube,Node);
	  }
	else
	  assert(0);
      }
  }
}
      
void cuddBddExistAbstractRecur(manager, f, cube, from, caller)
DdManager *manager;
DdNodePar  f;
DdNodePar  cube;
int from;
DdCallTreeNode caller;
{
  DdNodePar t,e,T,E,res,t_cube,e_cube;
  DdCallTreeNode then_child,Node;
  DdCacheHandle then_handle,else_handle,handle;
  int t_found=FALSE,e_found=FALSE;

  /* assumptions: f not constant, cube not constant, f.index<=cube.index */ 

  /* Compute the cofactors of f. */
  T = Cudd_T_Par(manager,f); 
  E = Cudd_E_Par(manager,f); 
  if(Cudd_IsComplementPar(f)) {
    T=Cudd_NotPar(T);
    E=Cudd_NotPar(E);
  }

  t_cube=cube;
  t=cuddBddExistsSimple(manager,T,&t_cube,&then_handle);
  e_cube=cube;
  e=cuddBddExistsSimple(manager,E,&e_cube,&else_handle);
  
  t_found=!cuddEqualPar(t,CUDD_WAIT);
  e_found=!cuddEqualPar(e,CUDD_WAIT);
  
  if(t_found && e_found)
    {	    
      /* If the two indices are the same, so are their levels. */
      if (f.index == cube.index) {
	t=Cudd_NotPar(t);
	e=Cudd_NotPar(e);
	res=cuddBddAndBasePar(manager,t,e);
	if(!cuddEqualPar(res,CUDD_ERROR))
	  {
	    res=Cudd_NotPar(res);
	    if(from==manager->Mypos)
	      DdCallTreeUpdateNode(manager,caller,res);
	    else
	      DdMessageReply(manager,from,res,caller);	  
	  }
	else
	  {
	    handle = cuddCacheCheck2Par(manager, CUDD_BDD_AND, t, e);
	    if(cuddCacheStatusPar(manager,then_handle)==CUDD_CACHE_HIT)
	      {
		res=Cudd_NotPar(cuddCacheResult(manager,handle));
		if(from==manager->Mypos)
		  DdCallTreeUpdateNode(manager,caller,res);
		else
		  DdMessageReply(manager,from,res,caller);	  
	      }
	    else
	      {
		Node=DdCallTreeCreateNode_PropNeg(manager,caller,from,handle);
		if(cuddCacheStatusPar(manager,handle)==CUDD_CACHE_IN_PROGRESS)
		  cuddCacheWait(manager,handle,Node);
		else if(cuddCacheStatusPar(manager,then_handle)==CUDD_CACHE_MISS)
		  DdWorkCreate2(manager,CUDD_BDD_AND,t,e,Node);
		else
		  assert(0);
	      }
	  }
      } else { /* if (f.index < cube.index) */
	res=cuddUniqueInterSafePar(manager,f.index,t,e);
	if(from==manager->Mypos)
	  DdCallTreeUpdateNode(manager,caller,res);
	else
	  DdMessageReply(manager,from,res,caller);	  
      }
    }
  else
    {
      /* note the handle does not a veldefined value if the result
	 as already available */
      
      /* If the two indices are the same, so are their levels. */
      if (f.index == cube.index) 
	{
	  then_child=DdCallTreeCreateNode_Exists(manager,caller,from,
						 then_handle,t,else_handle,e);
	} 
      else 
	{ /* if (f.index < cube.index) */  
	  then_child=DdCallTreeCreateNode(manager,caller,from,f.index,
					  then_handle,t,else_handle,e);
	}
      
      if(!e_found)
	{
	  if(cuddCacheStatusPar(manager,else_handle)==CUDD_CACHE_IN_PROGRESS)
	    {
	      cuddCacheWait(manager,else_handle,DdCallTreeNodeOther(then_child));
	    }
	  else if(cuddCacheStatusPar(manager,else_handle)==CUDD_CACHE_MISS)
	    DdWorkCreate2(manager,CUDD_BDD_EXISTS,E,e_cube,
			  DdCallTreeNodeOther(then_child));
	  else
	    assert(0);
	} 
      
      if(!t_found)
	{
	  if(cuddCacheStatusPar(manager,then_handle)==CUDD_CACHE_IN_PROGRESS)
	    {
	      cuddCacheWait(manager,then_handle,then_child);
	    }
	  else if(cuddCacheStatusPar(manager,then_handle)==CUDD_CACHE_MISS)
	    DdWorkCreate2(manager,CUDD_BDD_EXISTS,T,t_cube,
			  then_child);
	  else
	    assert(0);
	}      
    }

} /* end of cuddBddExistAbstractRecur */


static DdNodePar cuddBddExistsSimple(DdManager *manager,DdNodePar f, DdNodePar *cube, DdCacheHandle *handle)
{

  if(cuddIsConstantPar(f))
    {
      return f;
    }

  /* Abstract a variables that does not appear in f */
  if(!cuddLocal(manager,f.index))
    { /* remove all the local variables from cube and pass on the work */
      while(cube->index!=CUDD_MAXINDEX && cuddLocal(manager,cube->index))
	*cube=Cudd_T_Par(manager,*cube);
    }
  else 
    {
      while(f.index>cube->index)
	*cube=Cudd_T_Par(manager,*cube);
    }

  if(cuddEqualPar(*cube,DD_ONE_PAR(manager)))
    /* f does not depend on any of the variables */
    return f;

  *handle = cuddCacheCheck2Par(manager, CUDD_BDD_EXISTS, f, *cube);
  if(cuddCacheStatusPar(manager,*handle)==CUDD_CACHE_HIT)
    {
      return cuddCacheResult(manager,*handle);
    }

  return CUDD_WAIT;
}

/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis [Checks whether cube is an BDD representing the product of
  positive literals.]

  Description [Returns 1 in case of success; 0 otherwise.]

  SideEffects [None]

******************************************************************************/
static int
bddCheckPositiveCube(manager, cube)
DdManager *manager;
DdNodePar cube;
{
    DdNodePar zero;
    if (Cudd_IsComplementPar(cube)) return(0);
    if (cuddEqualPar(cube,DD_ONE_PAR(manager))) return(1);
    if (cuddIsConstantPar(cube)) return(0);
    zero=Cudd_NotPar(DD_ONE_PAR(manager));
    if (cuddEqualPar(Cudd_E_Par(manager,cube),zero)) {
        return(bddCheckPositiveCube(manager, Cudd_T_Par(manager,cube)));
    }
    return(0);

} /* end of bddCheckPositiveCube */
