#include "cuddInt.h"
#include "message.h"

#define PrintNode(node) \
  fprintf(fp,"\"%d,%d,%d\";\n",node.index,node.address.page,node.address.number);

#define PrintNodeElse(node) \
  fprintf(fp,"\"%d,%d,%d\" [style=%s];\n",node.index,node.address.page,abs(node.address.number),node.address.number>0 ? "dashed" : "dotted");

/**Function********************************************************************

  Synopsis    [Writes a dot file representing the argument DDs.]

  Description [Writes a file representing the argument DDs in a format
  suitable for the graph drawing program dot.
  It returns 1 in case of success; 0 otherwise (e.g., out-of-memory,
  file system full).
  Cudd_DumpDot does not close the file: This is the caller
  responsibility. Cudd_DumpDot uses a minimal unique subset of the
  hexadecimal address of a node as name for it.
  If the argument inames is non-null, it is assumed to hold the pointers
  to the names of the inputs. Similarly for onames.
  Cudd_DumpDot uses the following convention to draw arcs:
    <ul>
    <li> solid line: THEN arcs;
    <li> dotted line: complement arcs;
    <li> dashed line: regular ELSE arcs.
    </ul>
  The dot options are chosen so that the drawing fits on a letter-size
  sheet.
  ]

  SideEffects [None]

  SeeAlso     [Cudd_DumpBlif Cudd_PrintDebug]

******************************************************************************/
/*
this has changed but provided for backwards compatability
all references user nodes and the intire BDD is always printed
dot will be written in DD_DOT_FILE (should be defined in makefile)
*/

int
Cudd_DumpDot(dd,n,f,inames,onames,fp)
DdManager *dd;	/* manager */
int n;		/* number of output nodes to be dumped */
DdNode **f;	/* array of output nodes to be dumped */
char **inames;	/* array of input names (or NULL) */
char **onames;	/* array of output names (or NULL) */
FILE *fp;	/* pointer to the dump file */
{
#ifdef DD_DOT_FILE
  return(cuddDumpDotMaster(dd,inames,onames,DD_DOT_FILE));
#else
  return(0);
#endif
}

int Cudd_DumpDotNew(dd,inames,onames,file_name)
DdManager *dd;	/* manager */
char **inames;	/* array of input names (or NULL) */
char **onames;	/* array of output names (or NULL) */
char *file_name; /* the dump file */
{
  return(cuddDumpDotMaster(dd,inames,onames,file_name));
}

int
cuddDumpDotMaster2(manager,inames,onames,fp)
DdManager *manager;	/* manager */
char **inames;	/* array of input names (or NULL) */
char **onames;	/* array of output names (or NULL) */
FILE *fp;
{
  int i;
  DdNode *node;
  int retval;
  int ctr;

  /* Write the header and the global attributes. */
  retval = fprintf(fp,"digraph \"DD\" {\n");
  if (retval == EOF) return(0);
  retval = fprintf(fp,
		   "size = \"7.5,10\"\ncenter = true;\nedge [dir = none];\n");
  if (retval == EOF) return(0);

  /* Write the input name subgraph by scanning the support array. */
  retval = fprintf(fp,"{ node [shape = plaintext];\n");
  if (retval == EOF) return(0);
  retval = fprintf(fp,"  edge [style = invis];\n");
  if (retval == EOF) return(0);
  /* We use a name ("CONST NODES") with an embedded blank, because
  ** it is unlikely to appear as an input name.
  */
  retval = fprintf(fp,"  \"CONST NODES\" [style = invis];\n");
  if (retval == EOF) return(0);
  for (i = 0; i < manager->size; i++) {
    retval = fprintf(fp,"\" %d \" ",i);
    if (inames != NULL) {
      fprintf(fp,"[label=\"%s\"];\n",inames[manager->InvOrigPerm[i]]);
    } else {
      fprintf(fp,"[label=\"%d\"];\n",manager->InvOrigPerm[i]);
    }
    if (retval == EOF) return(0);
  }
  for (i = 0; i < manager->size; i++) {
    retval = fprintf(fp,"\" %d \" -> ", i);
    if (retval == EOF) return(0);
  }
  retval = fprintf(fp,"\"CONST NODES\"; \n}\n");
  if (retval == EOF) return(0);
  
  /* Write the output node subgraph. */
  retval = fprintf(fp,"{ rank = same; node [shape = box]; edge [style = invis];\n");
  if (retval == EOF) return(0);

  ctr=0;
  for (i = 0; i < manager->UserNodes.slots; i++) 
    {
      node = manager->UserNodes.nodelist[i];
      while (node!=NULL)
	{
	  if(node->ref>0 || !manager->gcEnabled)
	    {
	      if (onames == NULL) {
		retval = fprintf(fp,"\"%x\"; ", (int)node);
	      } else {
		retval = fprintf(fp,"\"  %s  \"; ", onames[ctr++]);
	      }
	      if (retval == EOF) return(0);	      
	    }	    
	  node=node->next;
	}
    }


  retval = fprintf(fp,"}\n");
  if (retval == EOF) return(0);
  /* note no invisible edges between user nodes 
  for (i = 0; i < n; i++) {
     if (i == n - 1) {
       retval = fprintf(fp,"; }\n");
     } else {
       retval = fprintf(fp," -> ");
     }
     if (retval == EOF) return(0);
  }
  */
  
  /* constant one node */
  retval = fprintf(fp,"{ node [shape = box];  edge [style = invis]; \"%d,%d,%d\" [label = \"1\"];}\n",manager->one.index,manager->one.address.page,manager->one.address.number);
  if (retval == EOF) return(0);
  retval = fprintf(fp,"{rank=same;  \"CONST NODES\"; \"%d,%d,%d\";}\n",manager->one.index,manager->one.address.page,manager->one.address.number);
  if (retval == EOF) return(0);

  /* user node edges */
  ctr=0;
  for (i = 0; i < manager->UserNodes.slots; i++) 
    {
      node = manager->UserNodes.nodelist[i];
      while (node!=NULL)
	{
	  if(node->ref>0 || !manager->gcEnabled)
	    {
	      if (onames == NULL) {
		retval = fprintf(fp,"\"%x\" ", (int)node);
	      } else {
		retval = fprintf(fp,"\"  %s  \" ", onames[ctr++]);
	      }
	      if (retval == EOF) return(0);	      
	      retval = fprintf(fp," -> ");
	      if (retval == EOF) return(0);
	      retval = PrintNode(node->NodePar);
	      if (retval == EOF) return(0);	      
	    }	    
	  node=node->next;
	}
    }

  return(1);
}
