/**CHeaderFile*****************************************************************

  FileName    [cuddInt.h]

  PackageName [cudd]

  Synopsis    [Internal data structures of the CUDD package.]

  Description []

  SeeAlso     []

  Author      [Fabio Somenzi]

  Copyright [This file was created at the University of Colorado at
  Boulder.  The University of Colorado at Boulder makes no warranty
  about the suitability of this software for any purpose.  It is
  presented on an AS IS basis.]

  Revision    [$Id: cuddInt.h,v 1.75 1997/04/13 17:29:17 fabio Exp $]

  Rev2  This file was attacked on Nov 3rd to remove all Zdd and Add, and
        any other noncritical functionality as far as our project is concerned..
        Changes include:
            DdSubtable: nodelist is a ptr to a DdNodeOld ptr for the hash table
            DdManager: vars and stack are ptrs to DdNodePars
                       memoryList is a ptr to DdNodeOld ptr
                       nextFree is a ptr to a DdNodeOld
            macro cuddRefPar   KIM:16/11 almost nuked
            macro cuddDerefPar KIM:16/11 almost nuked
            macro cuddIsConstantPar
            macro cuddTPar
            macro cuddEPar
            macro DD_ONE_PAR
            various EXTERN declarations...  

  Nov16     Uncommented cache variables in manager
            Changed DdNodePars in cache to DdNodeOld*

******************************************************************************/

#ifndef _CUDDINT
#define _CUDDINT


/*---------------------------------------------------------------------------*/
/* Nested includes                                                           */
/*---------------------------------------------------------------------------*/

#ifdef DD_MIS
#include "array.h"
#include "list.h"
#include "st.h"
#include "espresso.h"
#include "node.h"
#ifdef SIS
#include "graph.h"
#include "astg.h"
#endif
#include "network.h"
#endif

#include <math.h>
#include "cudd.h"
#include "mtr.h"

#if defined(__GNUC__)
#define DD_INLINE __inline__
# if (__GNUC__ >2 || __GNUC_MINOR__ >=7)
#   define DD_UNUSED __attribute__ ((unused))
# else
#   define DD_UNUSED
# endif
#else
#define DD_INLINE
#define DD_UNUSED
#endif


/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/
 
/* #define DD_MAXREF		((DdHalfWord) ~0) */
#define DD_EXTERN_REF		((DdHalfWord) ~0)
#define DD_NEW_NODE             (DD_EXTERN_REF-1)
#define DD_MAXREF		(DD_EXTERN_REF-2)

#define DD_DEFAULT_RESIZE	10	/* how many extra variables */
					/* should be added when resizing */
#define DD_MEM_CHUNK		(1<<12)

/* These definitions work for CUDD_VALUE_TYPE == double */
#define DD_ONE_VAL		(1.0)
#define DD_ZERO_VAL		(0.0)
#define DD_EPSILON		(1.0e-12)

/* The definitions of +/- infinity in terms of HUGE_VAL work on
** the DECstations and on many other combinations of OS/compiler.
*/
#ifdef HAVE_IEEE_754
#  define DD_PLUS_INF_VAL	(HUGE_VAL)
#else
#  define DD_PLUS_INF_VAL	(10e301)
#  define DD_CRI_HI_MARK	(10e150)
#  define DD_CRI_LO_MARK	(-(DD_CRI_HI_MARK))
#endif
#define DD_MINUS_INF_VAL	(-(DD_PLUS_INF_VAL))

#define DD_NON_CONSTANT		((DdNode *) 1)	/* for Cudd_bddIteConstant */

/* Unique table and cache management constants */
#define DD_MAX_SUBTABLE_DENSITY 4	/* tells when to resize a subtable */
/* gc when this percent are dead (measured w.r.t. slots, not keys)
** The first limit (LO) applies normally. The second limit applies when
** the package believes more space for the unique table (i.e., more dead
** nodes) would improve performance, and the unique table is not already
** too large.
*/
#define DD_GC_PERCENT_LO	DD_MAX_SUBTABLE_DENSITY * 20
#define DD_GC_PERCENT_HI	DD_MAX_SUBTABLE_DENSITY * 90
#define DD_LOOSE_UP_TO		800000	/* slow growth when unique table
					   beyond this limit (default) */
#define DD_GROW_FACTOR		2	/* table increases on rehash */
#define DD_MIN_HIT		30	/* resize cache when hit ratio
					   above this percentage (default) */

/* Variable ordering constants */
#define DD_SIFT_MAX_VAR		1000
#define DD_SIFT_MAX_SWAPS	2000000
#define DD_DEFAULT_RECOMB	0
#define DD_MAX_REORDER_GROWTH	1.2
#define DD_FIRST_REORDER	4004	/* 4 for the constants */
#define DD_DYN_RATIO		2	/* when to dynamically reorder */

/* Rev3: new constant */
#define NULL_PAGE (DdHalfWord) ~0
#define NULL_NUMBER 0 

static DdNodeAdr NULL_ADR = {NULL_PAGE,NULL_NUMBER};
static DdNodePar CUDD_ERROR = {0,{NULL_PAGE,NULL_NUMBER}};
static DdNodePar CUDD_WAIT = {1,{NULL_PAGE,NULL_NUMBER}};
static DdNodePar CUDD_FORWARD = {2,{NULL_PAGE,NULL_NUMBER}};

#define NULL_CALL_TREE_NODE 0
#define DD_CALL_TREE_SIZE   1024
#define DD_CALL_TREE_GROWTH 2

#define DD_WORK_QUEUE_SIZE   1024
#define DD_WORK_QUEUE_GROWTH 2

#define DD_DEFAULT_PAGES        10 /* initial number of entries in pagetable */
#define DD_PAGE_GROWTH          2  /* growth rate of entries in pagetable */



#define CUDD_CACHE_MISS         0
#define CUDD_CACHE_HIT          1
#define CUDD_CACHE_IN_PROGRESS  2


/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/* Hooks in CUDD are functions that the application registers with the
** manager so that they are called at appropriate times. The functions
** are passed the manager as argument; they should return 1 if
** successful and 0 otherwise.
*/
typedef struct DdHook {		/* hook list element */
    int (*f)();			/* function to be called */
    struct DdHook *next;	/* next element in the list */
} DdHook;

#ifdef __osf__
#pragma pointer_size save
#pragma pointer_size short
#endif

/* placed here because cache waiting list needs it */
typedef int DdCallTreeNode;

/********************* Cache waiting list stuff */
typedef struct DdCacheWaitingList
{
  DdCallTreeNode caller;
  struct DdCacheWaitingList *next;
} *DdCacheWaitingList;

/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   cache handles waiting calls. nov 25
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

typedef struct DdCache {
    DdNodePar f, g, h;		/* DDs */
    DdNodePar data;		/* already constructed DD */
    int op;		        /* operator */
    /* NOTE using constants to represent operation */ 

    struct DdCacheWaitingList *WaitingList;
    /* waiting list has more to do with cortrol flow, but
       it is convinient to keep it with the cache entry */
    
#ifdef DD_CACHE_PROFILE
    int count;
#endif
} DdCache;

typedef struct DdCacheHandle
{
  int status;
  DdCache *entry;
} DdCacheHandle;

#define cuddCacheStatusPar(manager,handle) ((handle).status)
#define cuddCacheResult(manager,handle) ((handle).entry->data)
#define cuddCacheEntry(manager,handle) ((handle).entry)

extern void cuddCacheUpdatePar(DdManager *table, DdCache *entry, DdNodePar data);
extern DdCacheHandle cuddCacheCheck2Par(DdManager *table, int op, DdNodePar f, DdNodePar g);
extern DdCacheHandle cuddCacheCheck3Par(DdManager *table, int op, DdNodePar f, DdNodePar g,DdNodePar h);
extern DdCacheHandle cuddCacheCheckItePar(DdManager *table, int op, DdNodePar f, DdNodePar g,DdNodePar h);
extern void cuddCacheWait(DdManager *manager,DdCacheHandle handle,DdCallTreeNode caller);
extern void ddCacheResizePar(DdManager *table);

/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  Rev2 nodelist is a pointer to a DdNodeOld pointer for the hash table
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

typedef struct DdSubtable {	/* subtable for one index */
    DdNodeAdr *nodelist;	/* hash table */
    unsigned int slots;		/* size of the hash table */
    unsigned int keys;		/* number of nodes stored in this table */
    unsigned int maxKeys;	/* slots * DD_MAX_SUBTABLE_DENSITY */
    unsigned int next;		/* index of next variable in group */
    DdNodePar *ReorderNodelist;	/* hash table for use when reordering */
} DdSubtable;


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  Rev3 new memory structure
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

typedef struct DdMemory
{
  DdNodeOld **PageTable;        /* array of startadresses of pages */
  DdNodeMore **PageTable2;      /* other info on each node, sored elsewhere */
  DdHalfWord nextPage;          /* next page to be allocated */
  DdHalfWord TableSize;         /* current number of entries in pagetable */
  DdNodeAdr nextFree;	        /* list of free nodes */
  unsigned int allocated;       /* number allocated nodes */
} DdMemory;

#define DdAllocatedInMem(mem) (cuddEqualAdr((mem)->nextFree,NULL_ADR) ? (mem)->nextPage*DD_MEM_CHUNK : ((mem)->nextPage-1)*DD_MEM_CHUNK+(mem)->nextFree.number-1)

/*++++++++++++++++++++++++++++++ call tree ++++++++++++++++++++++++++++++*/

typedef struct DdCallTreeNode
{
  DdCallTreeNode Parent;
  int ParentProcess;
  DdHalfWord index;            /* index of the variable split on */
  DdHalfWord state;
  union {
    DdCache *CacheEntry;
    DdNodePar Result;
  } Then;
  union {
    DdCache *CacheEntry;
    DdNodePar Result;
  } Else;
} DdCallTreeNodeStruct;

typedef struct DdCallTree
{
  int Size;
  DdCallTreeNodeStruct *NodeList;
  DdCallTreeNode FreeList;
} DdCallTree;

/*++++++++++++++++++++++++++++++ work queue ++++++++++++++++++++++++++++++*/
typedef struct DdWork
{
  int op;
  int index;
  int FromProcess;
  DdNodePar f,g,h;
  DdCallTreeNode caller;
} DdWork;

typedef struct DdWorkQueue
{
  DdWork *Stack;
  int StackPointer;
  int Size;
} DdWorkQueue;

/*+++++++++++++++++++++++ User Node Hashtable ++++++++++++++++++++++++++++*/
typedef struct DdUserNodeTable {
  DdNode **nodelist;  	      /* hash table */
  unsigned int slots;	      /* size of the hash table */
  unsigned int keys;	      /* number of nodes stored in this table */
  unsigned int dead;          /* number of dead nodes (reorder only) */
  unsigned int maxKeys;	      /* slots * DD_MAX_SUBTABLE_DENSITY */
  /* memory management */
  DdNode **PageTable;         /* array of pages */
  DdHalfWord nextPage;        /* next page to be allocated */
  DdHalfWord TableSize;       /* current number of entries in pagetable */
  DdNode *nextFree;	      /* list of free nodes */
} DdUserNodeTable;


struct DdManager {	/* specialized DD symbol table */
    /* Constants */
    DdNodePar one;		/* constant 1 */
    DdNode *oneUser;

    /* Computed Table */

    DdCache *cache;	          /* the cache */
    unsigned int cacheSlots;	  /* total number of cache entries */
    unsigned int cacheLookUps;  /* number of cache look-ups */
    unsigned int cacheHits;	  /* number of cache hits */
    unsigned int minHit;	  /* hit percentage above which to resize */
    unsigned int maxCache;	  /* size above which no resizing occurs */
    unsigned int maxCacheHard;  /* max value for maxCache */
   
 /* Unique Table */
    int size;			/* number of unique subtables */
    int maxSize;		/* max number of subtables before resizing */
    DdSubtable *subtables;	/* array of unique subtables */
    unsigned int slots;		/* total number of hash buckets */
    unsigned int keys;		/* total number of BDD and ADD nodes */
    int gcEnabled;		/* gc is enabled */
    unsigned int looseUpTo;	/* slow growth beyond this limit */
				/* (measured w.r.t. slots, not keys) */
    int allocated;		/* number of nodes allocated */
    int *allocatedPerLevel;
    int dead;

    int isolated;		/* isolated projection functions */
    int *perm;			/* current variable perm. (index to level) */
    int *invperm;		/* current inv. var. perm. (level to index) */

    DdNode **vars;		/* projection functions */
    long *interact;		/* interacting variable matrix */
    long *linear;		/* linear transform matrix */
    int linearSize;		/* number of rows and columns of linear */
    DdNodePar *stack;		/* stack for iterative procedures */
    /* Free List */
    /* Rev3 NEW MEMORY CONCEPT */
    DdMemory *memory;	        /* array of memory managers for symbol table */
                                /* each index is managed seperatly */
    int reordering;		/* flag indicating reordering is in progress */
#ifdef DD_MIS
    /* mis/verif compatibility fields */
    array_t *iton;		/* maps ids in ddNode to node_t */
    array_t *order;		/* copy of order_list */
    lsHandle handle;		/* where it is in network BDD list */
    network_t *network;
    st_table *local_order;	/* for local BDDs */
    int nvars;			/* variables used so far */
    int threshold;		/* for pseudo var threshold value*/
#endif
    /* General Parameters */
    CUDD_VALUE_TYPE epsilon;	/* tolerance on comparisons */
    /* Dynamic Reordering Parameters */
    int reorderings;		/* number of calls to Cudd_ReduceHeap */
    int siftMaxVar;		/* maximum number of vars sifted */
    int siftMaxSwap;		/* maximum number of swaps per sifting */
    double maxGrowth;		/* maximum growth during reordering */
    int autoDyn;		/* automatic dynamic reordering flag (BDD) */
    int autoMethod;		/* default reordering method */
    unsigned int nextDyn;	/* reorder if this size is reached */
    MtrNode *tree;		/* Variable group tree (BDD) */
    int groupcheck;		/* Used during group sifting */
    int recomb;			/* Used during group sifting */
    int symmviolation;		/* Used during group sifting */
    int arcviolation;		/* Used during group sifting */
    int populationSize;		/* population size for GA */
    int	numberXovers;		/* number of crossovers for GA */
#ifdef __osf__
#pragma pointer_size restore
#endif
    char *hooks;		/* application-specific field (used by vis) */
    DdHook *preGCHook;		/* hooks to be called before GC */
    DdHook *postGCHook;		/* hooks to be called after GC */
    DdHook *postReorderingHook;	/* hooks to be called after reordering */
#ifdef __osf__
#pragma pointer_size save
#pragma pointer_size short
#endif
    /* Statistical counters. */
    long memused;		/* total memory allocated for the manager */
    int garbageCollections;	/* number of garbage collections */
    int totCachehits;		/* total number of cache hits */
    int cachecollisions;	/* number of cache collisions */
    int cacheinserts;		/* number of cache insertions */
    int cachedeletions;		/* number of deletions during garbage coll. */

    int *Local;                 /* array indicating what machine each level 
				   is stored */
    int Mypos;
    struct ProcessInfo *ProcessInfo;
    DdCallTree CallTree;
    DdWorkQueue WorkQueue;
    DdUserNodeTable UserNodes;
    int *OrigPerm;              /* current index of original indexes
				   also present in Vars */
    int *InvOrigPerm;           /* current index to original indexes */

    int *Profile;
    int *ReorderBlocks;
    DdNodePar ReorderFreeList;

    int ForwardNodes;
};

typedef struct Move {
    DdHalfWord x;
    DdHalfWord y;
    unsigned int flags;
    int size;
    struct Move *next;
} Move;

#ifdef __osf__
#pragma pointer_size restore
#endif

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  Rev 3: memory Macros
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define cuddNextPar(unique,node) ((unique)->memory[(node).index].PageTable[(node).address.page][abs((node).address.number)].next)

#define cuddReorderNextIndex(unique,node) ((unique)->memory[(node).index].PageTable2[(node).address.page][abs((node).address.number)].Reorder.NextIndex)

#define cuddReorderIndex(unique,node) ((unique)->memory[(node).index].PageTable2[(node).address.page][abs((node).address.number)].Reorder.Index)

#define cuddReorderReadIndex(unique,node) (\
(!cuddLocal(unique,(node).index) || (node).index==CUDD_MAXINDEX) ? \
(node).index : \
(unique)->memory[(node).index].PageTable2[(node).address.page][abs((node).address.number)].Reorder.Index)

#define cuddPtrPar(unique,node) (&(unique)->memory[(node).index].PageTable[(node).address.page][abs((node).address.number)])



/**Macro***********************************************************************

  Synopsis    [Adds node to the head of the free list.]

  Description [Adds node to the head of the free list.  Does not
  deallocate memory chunks that become free.  This function is also
  used by the dynamic reordering functions.]

  SideEffects [None]

  SeeAlso     [cuddAllocNode cuddDynamicAllocNode]

******************************************************************************/

#ifdef DD_STATS
#define cuddDeallocNode(unique,node) \
    (node)->next = (unique)->nextFree; \
    (unique)->nextFree = node; \
    cudd_Nodesfreed++;
#else

#define cuddDeallocNode(unique,node) \
    (node)->next = (unique)->nextFree; \
    (unique)->nextFree = node;
#endif


/**Macro***********************************************************************

  Synopsis     [Increases the reference count of a node, if it is not
  saturated.]

  Description  [Increases the reference count of a node, if it is not
  saturated. This being a macro, it is faster than Cudd_Ref, but it
  cannot be used in constructs like cuddRef(a = b()).]

  SideEffects  [none]

  SeeAlso      [Cudd_Ref]

******************************************************************************/
/* Rev2 -- no ref counts ??? 
#define cuddRef(n) cuddSatInc(Cudd_Regular(n)->ref)

#define cuddRefPar(nodePar) (nodePar) 
*/
/**Macro***********************************************************************

  Synopsis     [Decreases the reference count of a node, if it is not
  saturated.]

  Description  [Decreases the reference count of node. It is primarily
  used in recursive procedures to decrease the ref count of a result
  node before returning it. This accomplishes the goal of removing the
  protection applied by a previous cuddRef. This being a macro, it is
  faster than Cudd_Deref, but it cannot be used in constructs like
  cuddDeref(a = b()).]

  SideEffects  [none]

  SeeAlso      [Cudd_Deref]

******************************************************************************/
/* Rev 2 -- no ref counts
#define cuddDeref(n) cuddSatDec(Cudd_Regular(n)->ref)

#define cuddDerefPar(nodePar) (nodePar)
*/

/**Macro***********************************************************************

  Synopsis     [Returns 1 if the node is a constant node.]

  Description  [Returns 1 if the node is a constant node (rather than an
  internal node). All constant nodes have the same index
  (CUDD_MAXINDEX). The pointer passed to cuddIsConstant must be regular.]

  SideEffects  [none]

  SeeAlso      [Cudd_IsConstant]

******************************************************************************/
/*
#define cuddIsConstant(node) ((node)->index == CUDD_MAXINDEX)
*/

#define cuddIsConstantPar( nodePar ) ( nodePar.index == CUDD_MAXINDEX )

/**Macro***********************************************************************

  Synopsis     [Returns the then child of an internal node.]

  Description  [Returns the then child of an internal node. If
  <code>node</code> is a constant node, the result is unpredictable.
  The pointer passed to cuddT must be regular.]

  SideEffects  [none]

  SeeAlso      [Cudd_T]

******************************************************************************/
/*
#define cuddT(node) ((node)->type.kids.T)
*/

#define cuddTPar(unique,node) ((unique)->memory[(node).index].PageTable[(node).address.page][(node).address.number].type.kids.T)

/**Macro***********************************************************************

  Synopsis     [Returns the else child of an internal node.]

  Description  [Returns the else child of an internal node. If
  <code>node</code> is a constant node, the result is unpredictable.
  The pointer passed to cuddE must be regular.]

  SideEffects  [none]

  SeeAlso      [Cudd_E]

******************************************************************************/
/*
#define cuddE(node) ((node)->type.kids.E)
*/


#define cuddEPar(unique,node) ((unique)->memory[(node).index].PageTable[(node).address.page][(node).address.number].type.kids.E)



/**Macro***********************************************************************

  Synopsis     [Returns the value of a constant node.]

  Description  [Returns the value of a constant node. If
  <code>node</code> is an internal node, the result is unpredictable.
  The pointer passed to cuddV must be regular.]

  SideEffects  [none]

  SeeAlso      [Cudd_V]

******************************************************************************/
/*
#define cuddV(node) ((node)->type.value)
*/
#define cuddVPar(unique,node) ((unique)->memory[(node).index].PageTable[(node).address.page][(node).address.number].type.value)


/**Macro***********************************************************************

  Synopsis    [Finds the current position of variable index in the
  order.]

  Description [Finds the current position of variable index in the
  order.  This macro duplicates the functionality of Cudd_ReadPerm,
  but it does not check for out-of-bounds indices and it is more
  efficient.]

  SideEffects [none]

  SeeAlso     [Cudd_ReadPerm]

******************************************************************************/
/*
#define	cuddI(dd,index) (((index)==CUDD_MAXINDEX)?(int)(index):(dd)->perm[(index)])
*/

/**Macro***********************************************************************

  Synopsis    [Hash function for the unique table.]

  Description []

  SideEffects [none]

  SeeAlso     [ddCHash ddCHash2]

******************************************************************************/
#define ddHashPar(f,g,m) ddHash(((f).address.page<<10) + (unsigned)(f).address.number, ((g).address.page<<10) +(unsigned)(g).address.number, m) 

#ifndef DD_NOIDIV
#if SIZEOF_VOID_P == 8 && SIZEOF_INT == 4
#define ddHash(a,b,m) (((unsigned)((unsigned long)(a)) + \
			((unsigned)((unsigned long)(b)) << 5)) % (m))
#else
#define ddHash(a,b,m) (((unsigned)(a) + ((unsigned)(b) << 5)) % (m))
#endif
#else
#if SIZEOF_VOID_P == 8 && SIZEOF_INT == 4
#define ddHash(a,b,m) \
	(((unsigned)((unsigned long)(a)) + \
	  ((unsigned)((unsigned long)(b)) << 5)) - \
	(unsigned) ((unsigned)((double) \
			       ((unsigned)((unsigned long)(a)) + \
				((unsigned)((unsigned long)(b)) << 5)) / \
			       (double) m) * (double) m))
#else
#define ddHash(a,b,m) \
	(((unsigned)(a) + ((unsigned)(b) << 5)) - \
	(unsigned) ((unsigned)((double) \
			       ((unsigned)(a) + ((unsigned)(b) << 5)) / \
			       (double) m) * (double) m))
#endif
#endif

/**Macro***********************************************************************

  Synopsis    [Hash function for the user table.]

  Description []

  SideEffects [none]

  SeeAlso     [ddHash ddHashPar]

  NOTE NOIDIV ignored expect compiler to handle % effectively
******************************************************************************/
#if SIZEOF_VOID_P == 8 && SIZEOF_INT == 4
#define ddHashUser(f,m) (((unsigned)((unsigned long)(f).index) + \
			 ((unsigned)((unsigned long)(f).address.page) << 7 )+ \
			 ((unsigned)((unsigned long)(f).address.number) << 5))\
			  % (m))
#else
#define ddHashUser(f,m) (((unsigned)((f).index) + \
			 ((unsigned)((f).address.page) << 7) + \
			 ((unsigned)((f).address.number) << 5))\
			  % (m))
#endif

/**Macro***********************************************************************

  Synopsis    [Hash function for the cache.]

  Description []

  SideEffects [none]

  SeeAlso     [ddHash ddCHash2]

******************************************************************************/
#ifndef DD_NOIDIV
#if SIZEOF_VOID_P == 8 && SIZEOF_INT == 4
#define ddCHash(o,a,b,c,m) ((((unsigned)((unsigned long)(a)) << 11) + \
			     ((unsigned)((unsigned long)(b)) << 5) + \
			     ((unsigned)((unsigned long)(c)) << 7) + \
			     (unsigned)((unsigned long)(o))) % (m))
#else
#define ddCHash(o,a,b,c,m) ((((unsigned)(a) << 11) + ((unsigned)(b) << 5) + \
			     ((unsigned)(c) << 7) + (unsigned)(o)) % (m))
#endif
#else
#if SIZEOF_VOID_P == 8 && SIZEOF_INT == 4
#define ddCHash(o,a,b,c,m) \
	((((unsigned)((unsigned long)(a)) << 11) + \
	  ((unsigned)((unsigned long)(b)) << 5) + \
	  ((unsigned)((unsigned long)(c)) << 7) + \
	  (unsigned)((unsigned long)(o))) - (unsigned) \
	 ((unsigned)((double)(((unsigned)((unsigned long)(a)) << 11) + \
			      ((unsigned)((unsigned long)(b)) << 5) + \
			      ((unsigned)((unsigned long)(c)) << 7) + \
			      (unsigned)((unsigned long)(o))) / \
		     (double) m) * (double) m))
#else
#define ddCHash(o,a,b,c,m) \
	((((unsigned)(a) << 11) + ((unsigned)(b) << 5) + \
	  ((unsigned)(c) << 7) + (unsigned)(o)) - (unsigned) \
	 ((unsigned)((double)(((unsigned)(a) << 11) + \
			      ((unsigned)(b) << 5) + \
			      ((unsigned)(c) << 7) + \
			      (unsigned)(o)) / \
		     (double) m) * (double) m))
#endif
#endif


/**Macro***********************************************************************

  Synopsis    [Hash function for the cache for functions with two
  operands.]

  Description []

  SideEffects [none]

  SeeAlso     [ddHash ddCHash]

******************************************************************************/
#ifndef DD_NOIDIV
#if SIZEOF_VOID_P == 8 && SIZEOF_INT == 4
#define ddCHash2(o,a,b,m) ((((unsigned)((unsigned long)(a)) << 11) + \
			    ((unsigned)((unsigned long)(b)) << 5) + \
			    (unsigned)((unsigned long)(o))) % (m))
#else
#define ddCHash2(o,a,b,m) ((((unsigned)(a) << 11) + ((unsigned)(b) << 5) + \
			    (unsigned)(o)) % (m))
#endif
#else
#if SIZEOF_VOID_P == 8 && SIZEOF_INT == 4
#define ddCHash2(o,a,b,m) \
	((((unsigned)((unsigned long)(a)) << 11) + \
	  ((unsigned)((unsigned long)(b)) << 5) + \
	  (unsigned)((unsigned long)(o))) - \
	 (unsigned) ((unsigned)((double) \
				(((unsigned)((unsigned long)(a)) << 11) + \
				 ((unsigned)((unsigned long)(b)) << 5) + \
				 (unsigned)((unsigned long)(o))) / \
				(double) m) * (double) m))
#else
#define ddCHash2(o,a,b,m) \
	((((unsigned)(a) << 11) + ((unsigned)(b) << 5) + (unsigned)(o)) - \
	(unsigned) ((unsigned)((double)(((unsigned)(a) << 11) + \
					((unsigned)(b) << 5) + \
					(unsigned)(o)) / \
			       (double) m) * (double) m))
#endif
#endif


/**Macro***********************************************************************

  Synopsis    [Computes the minimum of two numbers.]

  Description []

  SideEffects [none]

  SeeAlso     [ddMax]

******************************************************************************/
#define ddMin(x,y) (((y) < (x)) ? (y) : (x))


/**Macro***********************************************************************

  Synopsis    [Computes the maximum of two numbers.]

  Description []

  SideEffects [none]

  SeeAlso     [ddMin]

******************************************************************************/
#define ddMax(x,y) (((y) > (x)) ? (y) : (x))


/**Macro***********************************************************************

  Synopsis    [Computes the absolute value of a number.]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/
#define ddAbs(x) (((x)<0) ? -(x) : (x))


/**Macro***********************************************************************

  Synopsis    [Returns 1 if the absolute value of the difference of the two
  arguments x and y is less than e.]

  Description []

  SideEffects [none]

  SeeAlso     []

******************************************************************************/
#define ddEqualVal(x,y,e) (ddAbs((x)-(y))<(e))


/**Macro***********************************************************************

  Synopsis    [Saturating increment operator.]

  Description []

  SideEffects [none]

  SeeAlso     [cuddSatDec]

******************************************************************************/
#if SIZEOF_VOID_P == 8 && SIZEOF_INT == 4
#define cuddSatInc(x) ((x)++)
     /* kim: assume no overflow??? */
#else
#define cuddSatInc(x) ((x) += (x) != (DdHalfWord)DD_MAXREF)
#endif


/**Macro***********************************************************************

  Synopsis    [Saturating decrement operator.]

  Description []

  SideEffects [none]

  SeeAlso     [cuddSatInc]

******************************************************************************/
#if SIZEOF_VOID_P == 8 && SIZEOF_INT == 4
#define cuddSatDec(x) ((x)--)
#else
#define cuddSatDec(x) ((x) -= (x) != (DdHalfWord)DD_MAXREF)
#endif


/**Macro***********************************************************************

  Synopsis    [Returns the constant 1 node.]

  Description []

  SideEffects [none]

  SeeAlso     [DD_ZERO DD_PLUS_INFINITY DD_MINUS_INFINITY]

******************************************************************************/

#define DD_ONE_PAR(dd)		((dd)->one)

#define DD_ONE(dd) ((dd)->oneUser)

/**Macro***********************************************************************

  Synopsis    [Returns the arithmetic 0 constant node.]

  Description [Returns the arithmetic 0 constant node. This is different
  from the logical zero. The latter is obtained by
  Cudd_Not(DD_ONE(dd)).]

  SideEffects [none]

  SeeAlso     [DD_ONE Cudd_Not DD_PLUS_INFINITY DD_MINUS_INFINITY]

******************************************************************************/
/*
#define DD_ZERO(dd) ((dd)->zero)
*/

/**Macro***********************************************************************

  Synopsis    [Returns the plus infinity constant node.]

  Description []

  SideEffects [none]

  SeeAlso     [DD_ONE DD_ZERO DD_MINUS_INFINITY]

******************************************************************************/
     /*
#define DD_PLUS_INFINITY(dd) ((dd)->plusinfinity)
*/

/**Macro***********************************************************************

  Synopsis    [Returns the minus infinity constant node.]

  Description []

  SideEffects [none]

  SeeAlso     [DD_ONE DD_ZERO DD_PLUS_INFINITY]

******************************************************************************/
     /*
#define DD_MINUS_INFINITY(dd) ((dd)->minusinfinity)
*/

/**Macro***********************************************************************

  Synopsis    [Enforces DD_MINUS_INF_VAL <= x <= DD_PLUS_INF_VAL.]

  Description [Enforces DD_MINUS_INF_VAL <= x <= DD_PLUS_INF_VAL.
  Furthermore, if x <= DD_MINUS_INF_VAL/2, x is set to
  DD_MINUS_INF_VAL. Similarly, if DD_PLUS_INF_VAL/2 <= x, x is set to
  DD_PLUS_INF_VAL.  Normally this macro is a NOOP. However, if
  HAVE_IEEE_754 is not defined, it makes sure that a value does not
  get larger than infinity in absolute value, and once it gets to
  infinity, stays there.  If the value overflows before this macro is
  applied, no recovery is possible.]

  SideEffects [none]

  SeeAlso     []

******************************************************************************/
#ifdef HAVE_IEEE_754
#  define cuddAdjust(x)
#else
#  define cuddAdjust(x)		((x) = ((x) >= DD_CRI_HI_MARK) ? DD_PLUS_INF_VAL : (((x) <= DD_CRI_LO_MARK) ? DD_MINUS_INF_VAL : (x)))
#endif


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Function prototypes                                                       */
/*---------------------------------------------------------------------------*/


EXTERN DdNodePar cuddBddAndRecur ARGS((DdManager *manager, DdNodePar f, DdNodePar g));
EXTERN DdNodePar cuddBddXorRecur ARGS((DdManager *manager, DdNodePar f, DdNodePar g));

EXTERN int cuddInitCachePar ARGS((DdManager *unique, unsigned int cacheSize, unsigned int maxCacheSize));
EXTERN void cuddCacheInsertPar ARGS((DdManager *table, int op, DdNodePar f, DdNodePar g, DdNodePar h, DdNodePar data));
EXTERN void cuddCacheInsert2Par ARGS((DdManager *table, int op, DdNodePar f, DdNodePar g, DdNodePar data));
EXTERN DdNodePar  cuddCacheLookupPar ARGS((DdManager *table, int op, DdNodePar f, DdNodePar g, DdNodePar h));
EXTERN DdNodePar  cuddCacheLookup2Par ARGS((DdManager *table, int op, DdNodePar f, DdNodePar g));
EXTERN void cuddCacheProfilePar ARGS((DdManager *table));
EXTERN void cuddPrintNode ARGS((DdNodePar f));
EXTERN void cuddGetBranches ARGS((DdNodePar g, DdNodePar *g1, DdNodePar *g0));

EXTERN DdNodePar cuddBddComposeRecur ARGS((DdManager *dd, DdNodePar f, DdNodePar g, DdNodePar proj));

EXTERN DdNodePar cuddSolveEqnRecur ARGS((DdManager *bdd, DdNodePar F, DdNodePar Y, DdNodePar *G, int n, int *yIndex, int i));
EXTERN DdNodePar cuddVerifySol ARGS((DdManager *bdd, DdNodePar F, DdNodePar *G, int *yIndex, int n));
#ifdef ST_INCLUDED
EXTERN DdNodePar cuddSplitSetRecur ARGS((DdManager *manager, st_table *mtable, int *varSeen, DdNodePar p, double n, double max, int index));
#endif
EXTERN DdNodeAdr cuddAllocNodePar ARGS((DdManager *unique,int index));
EXTERN DdManager * cuddInitTablePar ARGS(());
EXTERN void cuddFreeTable ARGS((DdManager *unique));
EXTERN int cuddGarbageCollectPar ARGS((DdManager *unique, int clearCache));
EXTERN DdNodePar cuddUniqueInterPar ARGS((DdManager *unique, int index, DdNodePar T, DdNodePar E));
EXTERN DdNodePar cuddUniqueInterSafePar ARGS((DdManager *unique, int index, DdNodePar t, DdNodePar e));

EXTERN int cuddInsertSubtables ARGS((DdManager *unique, int n, int level));
EXTERN int cuddDestroySubtables ARGS((DdManager *unique, int n));

EXTERN int cuddP ARGS((DdManager *dd, DdNodePar f));
#ifdef ST_INCLUDED
EXTERN enum st_retval cuddStCountfree ARGS((char *key, char *value, char *arg));
EXTERN int cuddCollectNodes ARGS((DdNodePar f, st_table *visited));
#endif

EXTERN DdNodePar cuddBddIsop ARGS((DdManager *dd, DdNodePar L, DdNodePar U));

/**AutomaticEnd***************************************************************/

#define cuddValuePar(f) ((f.index << 10) + (f.address.page << 8) + f.address.number)

#define cuddLevelPar(man,f) (f.index==CUDD_MAXINDEX ? CUDD_MAXINDEX : man->perm[f.index])
#define cuddLevelToIndex(man,level) ((man)->invperm[level])
#define cuddIndexToLevel(man,index) ((man)->perm[index])
#define cuddSetNewLevel(manager,index,level) ((manager)->perm[(manager)->invperm[level]=(index)]=(level))

/* order function for adresses (must be strict order) */
#define cuddAdrOrder(a,b) ((a).page==(b).page ? (a).number<(b).number : (a).page<(b).page)

/* Methods for CallTree (cuddCallTree.c) */
#define CUDD_CALL_TREE_DONE   0
#define CUDD_RETURN_ON        1
#define CUDD_RETURN_ON_THEN   1
#define CUDD_RETURN_ON_ELSE   2
#define CUDD_RETURN_ON_BOTH   3
#define CUDD_PROPAGATE        4
#define CUDD_PROPAGATE_NEGATE 5  
#define CUDD_OR_ON            6
#define CUDD_OR_ON_THEN       6
#define CUDD_OR_ON_ELSE       7
#define CUDD_OR_ON_BOTH       8

int DdCallTreeCreate(DdManager *manager);
DdCallTreeNode DdCallTreeCreateNode(DdManager *manager,
				    DdCallTreeNode Parent,
				    int ParentProcess,
				    int index,
				    int type);
int DdCallTreeNodeSetCacheEntry(DdManager *manager,DdCallTreeNode Node,DdCache *entry);
int DdCallTreeUpdateNode(DdManager *manager,DdCallTreeNode Node,DdNodePar Res);
#define DdCallTreeNodeOther(Node) (-(Node))
void DdCallTreeDestroy(DdManager *manager);

#define DdCallTreeCreateNodeNormal(man,P,PP,index) \
DdCallTreeCreateNode(man,P,PP,index,CUDD_RETURN_ON_BOTH)
#define DdCallTreeCreateNodeOr(man,P,PP) \
DdCallTreeCreateNode(man,P,PP,0,CUDD_OR_ON_BOTH)
#define DdCallTreeCreateNode_Propagate(man,P,PP)	\
DdCallTreeCreateNode(man,P,PP,0,CUDD_PROPAGATE)


#define cuddLocal(man,index) ((man)->Local[index]==(man)->Mypos)
#define cuddSetLocal(man,index) ((man)->Local[index]=(man)->Mypos)
#define cuddSetOwner(man,index,owner) ((man)->Local[index]=owner)
#define DdFindOwner(manager,index) (manager->Local[index])


/* CuddWork.c */
#define DdWorkQueuePush2(manager,op,index,process,f,g,caller) \
          DdWorkQueuePush(manager,op,index,process,f,g,CUDD_ERROR,caller)

int DdWorkQueuePush(DdManager *manager,int op,int index,int from_process,
DdNodePar f,DdNodePar g,DdNodePar h,DdCallTreeNode caller);

int DdWorkQueuePop(DdManager *manager,int *op,int *index,int *from_process,
DdNodePar *f,DdNodePar *g,DdNodePar *h,DdCallTreeNode *caller);

int DdWorkQueueCreate(DdManager *manager);

int DdWorkCreate2(DdManager *manager,int op,DdNodePar f,DdNodePar g,DdCallTreeNode caller);

int DdWorkCreate3(DdManager *manager,int op,DdNodePar f,DdNodePar g,DdNodePar h,DdCallTreeNode caller);

int DdWorkCreateIte(DdManager *manager,int op,DdNodePar f,DdNodePar g,DdNodePar h,DdCallTreeNode caller);

/* cuddTable.c */
extern void cuddClearSubtable(DdManager *manager,int index);
extern int CuddInitIndex(DdManager *manager,int index);
extern int ddResizeTablePar ARGS((DdManager *unique, int index));

/* cuddUserNode.c */
DdNodePar cuddNodeUserToPar(DdManager *manager,DdNode *node);
int cuddInitUserNodeTable(DdManager *manager);
DdNode* cuddUserNodeInter(DdManager *manager,DdNodePar node_par);
void cuddComputeOrigPerm(DdManager *manager);
DdNode* cuddAllocNodeUser(DdManager *manager);
int cuddDeallocNodeUser(DdManager *manager,DdNode *node);
int cuddUserNodeRehash(DdManager *manager);

/* reorder stuff */
#define cuddIncreaseRef(man,node) ((node).index < CUDD_MAXINDEX ? ((man)->memory[(node).index].PageTable2[(node).address.page][abs((node).address.number)].Reorder.Ref<DD_MAXREF ? ++(man)->memory[(node).index].PageTable2[(node).address.page][abs((node).address.number)].Reorder.Ref: DD_MAXREF) : DD_MAXREF)

#define cuddDecreaseRef(man,node) ((node).index < CUDD_MAXINDEX ? ((man)->memory[(node).index].PageTable2[(node).address.page][abs((node).address.number)].Reorder.Ref<DD_MAXREF ? --(man)->memory[(node).index].PageTable2[(node).address.page][abs((node).address.number)].Reorder.Ref: DD_MAXREF) : DD_MAXREF)

#define cuddDecreaseTempRef(tmp,node) ((node).index < CUDD_MAXINDEX ? ((tmp)[(node).index][(node).address.page][abs((node).address.number)].Reorder.Ref<DD_MAXREF ? \
--(tmp)[(node).index][(node).address.page][abs((node).address.number)].Reorder.Ref \
: DD_MAXREF) : DD_MAXREF)

/* cuddReorderSlave.c */
int cuddReorder(DdManager *manager);
/* cuddReorder.c */
int cuddIncreaseRefSafe(DdManager *manager, DdNodePar f);
/* cuddReorderMaster.c */
#define Cudd_ReduceHeap(man,method,min) cuddReorderMaster(man)
int cuddReorderMaster(DdManager *manager);

/* Dot funcions */
/* cuddDotSalve.c */
int cuddDumpDotSlave2(DdManager *manager,FILE *fp);

/* base functions for recurtion (cuddBddIteBase.c) */
extern DdNodePar cuddBddIteBase(DdManager *manager,DdNodePar f,DdNodePar g,DdNodePar h,int *op);
extern DdNodePar cuddBddAndBasePar(DdManager *manager,DdNodePar f,DdNodePar g);
extern DdNodePar cuddBddXorBasePar(DdManager *manager,DdNodePar f,DdNodePar g);

/* recursive functions (cuddRecur.c) */


extern void
cuddBddIteRecurPar(DdManager *manager,
		   int index,
		   DdNodePar  f,
		   DdNodePar  g,
		   DdNodePar  h,
		   int from,
		   DdCallTreeNode caller);

extern void
cuddBddRecur2(DdManager *manager,
		int index,
		DdNodePar  f,
		DdNodePar  g,
		int from,
		DdCallTreeNode caller,
		int op,
		DdNodePar (*base)(DdManager *,DdNodePar,DdNodePar));

extern void cuddBddExistAbstractRecurPre(DdManager *manager,
					 DdNodePar  f,
					 DdNodePar  cube,
					 int from,
					 DdCallTreeNode caller);
extern void cuddBddExistAbstractRecur(DdManager *manager,
				      DdNodePar  f,
				      DdNodePar  cube,
				      int from,
				      DdCallTreeNode caller);

extern void cuddBddAndAbstractRecurPre(DdManager *manager,
					 DdNodePar  f,
					 DdNodePar  g,
					 DdNodePar  cube,
					 int from,
					 DdCallTreeNode caller);
extern void cuddBddAndAbstractRecur(DdManager *manager,
				    int index,
				      DdNodePar  f,
				      DdNodePar  g,
				      DdNodePar  cube,
				      int from,
				      DdCallTreeNode caller);

extern void cuddBddRecurProcess2(DdManager *manager,int op, DdNodePar f, DdNodePar g,DdCallTreeNode caller);
extern void cuddBddRecurProcess3(DdManager *manager,int op, DdNodePar f, DdNodePar g, DdNodePar h,DdCallTreeNode caller);

#endif /* _CUDDINT */


